/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Martin Miller (martin dot miller at student dot kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "FramedPointWithNormal.h"

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/LinkedCoordinate.h>
#include <VirtualRobot/LinkedCoordinate.h>
#include <VirtualRobot/VirtualRobot.h>

#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>

using namespace armarx;



FramedPointWithNormal::FramedPointWithNormal()
{
}

FramedPointWithNormal::FramedPointWithNormal(const FramedPointWithNormal& source) :
    IceUtil::Shared(source),
    PointWithNormalBase(source),
    FramedPointWithNormalBase(source)
{
}

FramedPointWithNormal::FramedPointWithNormal(const Eigen::Vector3f& position, const Eigen::Vector3f& normal, const std::string& frame, const std::string& agent) :
    PointWithNormal(position, normal)
{
    this->frame = frame;
    this->agent = agent;

}

FramedPointWithNormal::FramedPointWithNormal(const PointWithNormal& pointWithNormal, const std::string& frame, const std::string& agent) :
    FramedPointWithNormal(pointWithNormal.positionToEigen(), pointWithNormal.normalToEigen(), frame, agent)
{
}

FramedPointWithNormal::FramedPointWithNormal(Ice::Float px, ::Ice::Float py, ::Ice::Float pz, Ice::Float nx, ::Ice::Float ny, ::Ice::Float nz, const std::string& frame, const std::string& agent) :
    PointWithNormal(px, py, pz, nx, ny, nz)
{
    this->frame = frame;
    this->agent = agent;
}

std::string FramedPointWithNormal::getFrame() const
{
    return frame;
}


void FramedPointWithNormal::changeFrame(const VirtualRobot::RobotPtr& robot, const std::string& newFrame)
{
    FramedPosition  framedPos(positionToEigen(), frame, agent);
    FramedDirection framedDir(normalToEigen(), frame, agent);

    framedPos.changeFrame(robot, newFrame);
    framedDir.changeFrame(robot, newFrame);
    this->px = framedPos.x;
    this->py = framedPos.y;
    this->pz = framedPos.z;

    this->nx = framedDir.x;
    this->ny = framedDir.y;
    this->nz = framedDir.z;
}

void FramedPointWithNormal::changeToGlobal(const SharedRobotInterfacePrx& referenceRobot)
{
    VirtualRobot::RobotPtr sharedRobot(new RemoteRobot(referenceRobot));
    changeFrame(sharedRobot, GlobalFrame);
}

void FramedPointWithNormal::changeToGlobal(const VirtualRobot::RobotPtr& referenceRobot)
{
    changeFrame(referenceRobot, GlobalFrame);
}

Eigen::Vector3f FramedPointWithNormal::positionToGlobalEigen(const SharedRobotInterfacePrx& referenceRobot) const
{
    return getFramedPosition().toGlobalEigen(referenceRobot);
}

Eigen::Vector3f FramedPointWithNormal::normalToGlobalEigen(const SharedRobotInterfacePrx& referenceRobot) const
{
    return getFramedNormal().toGlobalEigen(referenceRobot);
}

Eigen::Vector3f FramedPointWithNormal::positionToGlobalEigen(const VirtualRobot::RobotPtr& referenceRobot) const
{
    return getFramedPosition().toGlobalEigen(referenceRobot);
}

Eigen::Vector3f FramedPointWithNormal::normalToGlobalEigen(const VirtualRobot::RobotPtr& referenceRobot) const
{
    return getFramedNormal().toGlobalEigen(referenceRobot);
}


FramedPointWithNormalPtr FramedPointWithNormal::toRootFrame(const SharedRobotInterfacePrx& referenceRobot) const
{
    return new FramedPointWithNormal(positionToRootEigen(referenceRobot), normalToRootEigen(referenceRobot), frame, agent);
}

FramedPointWithNormalPtr FramedPointWithNormal::toRootFrame(const VirtualRobot::RobotPtr& referenceRobot) const
{
    return new FramedPointWithNormal(positionToRootEigen(referenceRobot), normalToRootEigen(referenceRobot), frame, agent);
}

Eigen::Vector3f FramedPointWithNormal::positionToRootEigen(const SharedRobotInterfacePrx& referenceRobot) const
{
    return getFramedPosition().toRootEigen(referenceRobot);
}

Eigen::Vector3f FramedPointWithNormal::normalToRootEigen(const SharedRobotInterfacePrx& referenceRobot) const
{
    return getFramedNormal().toRootEigen(referenceRobot);
}

Eigen::Vector3f FramedPointWithNormal::positionToRootEigen(const VirtualRobot::RobotPtr& referenceRobot) const
{
    return getFramedPosition().toRootEigen(referenceRobot);
}

Eigen::Vector3f FramedPointWithNormal::normalToRootEigen(const VirtualRobot::RobotPtr& referenceRobot) const
{
    return getFramedNormal().toRootEigen(referenceRobot);
}

FramedPosition FramedPointWithNormal::getFramedPosition() const
{
    return FramedPosition(positionToEigen(), frame, agent);
}

FramedDirection FramedPointWithNormal::getFramedNormal() const
{
    return FramedDirection(normalToEigen(), frame, agent);
}


std::string FramedPointWithNormal::output(const Ice::Current& c) const
{
    std::stringstream s;
    s << PointWithNormal::output(c) << std::endl << "frame: " << getFrame() << (agent.empty() ? "" : (" agent: " + agent));
    return s.str();
}


void FramedPointWithNormal::serialize(const ObjectSerializerBasePtr& serializer, const Ice::Current&) const
{
    AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);
    PointWithNormal::serialize(serializer);
    obj->setString("frame", frame);
    obj->setString("agent", agent);

}

void FramedPointWithNormal::deserialize(const ObjectSerializerBasePtr& serializer, const Ice::Current&)
{
    AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);
    PointWithNormal::deserialize(serializer);
    frame = obj->getString("frame");

    if (obj->hasElement("agent"))
    {
        agent = obj->getString("agent");
    }
}

